// database field container                                      -*- C++ -*-
// $Id: field.h,v 1.1 2004/05/22 17:58:07 roger Exp $
//
// Copyright (C) 2003  Roger Leigh <rleigh@debian.org>
//
//
//  All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions
//  are met:
//
//  * Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//  * Redistributions in binary form must reproduce the above
//    copyright notice, this list of conditions and the following
//    disclaimer in the documentation and/or other materials provided
//    with the distribution.
//  * Neither the name of the author, nor the names of other
//    contributors may be used to endorse or promote products derived
//    from this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
//  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
//  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
//  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS
//  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
//  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
//  ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
//  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
//  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
//  SUCH DAMAGE.
//
///////////////////////////////////////////////////////////////////////////

#ifndef PQXXOBJECT_FIELD_H
#define PQXXOBJECT_FIELD_H

#include <cassert>

#include <sigc++/object.h>
#include <sigc++/signal.h>

#include <pqxxobject/field_base.h>

namespace pqxxobject
{
  /**
   * Database field template class.
   * This class is used to represent a single field in a row of a
   * table.  This is a single value belonging to a column in a row,
   * rather than the whole column.  The value
   * may not be NULL.
   *
   * The class has some similarity with std::auto_ptr, but the copying
   * semantics are different: on copy, the value is copied, which may
   * involve memory allocation.  There is no transfer of pointer
   * ownership.  Contained objects must posess a copy constructor.
   *
   * As well as storing value, the class has the ability to emit
   * signals when the field value is changed.  Listeners (e.g. user
   * interface widgets) may connect to the signal and will receive
   * notification of changes as they occur.
   */
  template<typename T>
  class field : public pqxxobject::field_base
  {
  public:
    typedef T value_type;

    /// The constructor.
    explicit field(const value_type& value):
      m_value(value)
    {
      // No signal_changed on construction (signal handlers not yet
      // connected).
    }

    /// The copy constructor.
    field(const field<value_type>& rhs):
      field_base(rhs),
      m_value(rhs.m_value)
    {
      // No signal_changed on copy (signal handlers not yet
      // connected).
    }

    virtual ~field()
    {}

    /// Access member functions.
    const value_type *operator -> () const
    {
      return &m_value;
    }

    const value_type& operator * () const
    {
      return m_value;
    }

    /// Overloaded assignment operator.
    field<value_type>& operator = (const field<value_type>& rhs)
    {
      if (this != &rhs)
	set_value(rhs.m_value);
      return *this;
    }

    /// Overloaded assignment operator.
    field<value_type>& operator = (const value_type& rhs)
    {
      set_value(rhs);
      return *this;
    }

    field<value_type>& operator += (const field<value_type>& rhs)
    {
      m_value += rhs.m_value;
      m_signal_changed.emit();
      return *this;
    }

    field<value_type>& operator -= (const field<value_type>& rhs)
    {
      m_value -= rhs.m_value;
      m_signal_changed.emit();
      return *this;
    }

    field<value_type>& operator *= (const field<value_type>& rhs)
    {
      m_value *= rhs.m_value;
      m_signal_changed.emit();
      return *this;
    }

    field<value_type>& operator /= (const field<value_type>& rhs)
    {
      m_value /= rhs.m_value;
      m_signal_changed.emit();
      return *this;
    }

    field<value_type>& operator %= (const field<value_type>& rhs)
    {
      m_value %= rhs.m_value;
      m_signal_changed.emit();
      return *this;
    }

    field<value_type>& operator ++ ()
    {
      m_value++;
      m_signal_changed.emit();
      return *this;
    }

    field<value_type> operator ++ (int)
    {
      field<value_type> ret(*this);
      m_value++;
      m_signal_changed.emit();
      return ret;
    }

    field<value_type>& operator -- ()
    {
      m_value--;
      m_signal_changed.emit();
      return *this;
    }

    field<value_type> operator -- (int)
    {
      field<value_type> ret(*this);
      m_value--;
      m_signal_changed.emit();
      return ret;
    }

    friend field<value_type> operator + <> (const field<value_type>& lhs,
					    const field<value_type>& rhs);

    friend field<value_type> operator - <> (const field<value_type>& lhs,
					    const field<value_type>& rhs);

    friend field<value_type> operator * <> (const field<value_type>& lhs,
					    const field<value_type>& rhs);

    friend field<value_type> operator / <> (const field<value_type>& lhs,
					    const field<value_type>& rhs);

    friend field<value_type> operator % <> (const field<value_type>& lhs,
					    const field<value_type>& rhs);

    friend bool operator == <> (const field<value_type>& lhs,
				const field<value_type>& rhs);

    friend bool operator != <> (const field<value_type>& lhs,
				const field<value_type>& rhs);

    friend field<value_type> operator - <> (const field<value_type>& rhs);

    friend field<value_type> operator + <> (const field<value_type>& rhs);

    friend std::ostream& operator << <> (std::ostream& output_stream,
					 const field<value_type>& rhs);

//     friend std::istream& operator >> <> (std::istream& input_stream,
// 					 const field<value_type>& rhs);

    /// Conversion operator.
    operator const value_type& () const
    {
      return m_value;
    }

    /**
     * Get the contained value by reference.
     * @returns a reference to the value.
     */
    value_type& get_value()
    {
      return m_value;
    }

    /**
     * Get the contained value by constant reference.
     * @returns a constant reference to the value.
     */
    const value_type& get_value() const
    {
      return m_value;
    }

    /**
     * Set the contained value.
     * @param value the value to set.
     */
    void set_value(const value_type& value)
    {
      m_value = value;
      m_signal_changed.emit();
    }

    /**
     * Check if the contained value is NULL.
     * @returns true if NULL, otherwise false.
     */
    bool is_null() const
    {
      return false;
    }

    /**
     * Check if the contained value is not NULL.
     * @returns true if not NULL, otherwise false.
     */
    bool is_not_null() const
    {
      return true;
    }

  private:
    /// The contained value.
    value_type m_value;

  }; // class field


  template<typename T>
  inline field<T> operator + (const field<T>& lhs,
				       const field<T>& rhs)
  {
    return field<T>(lhs) += rhs;
  }

  template<typename T>
  inline field<T> operator - (const field<T>& lhs,
				       const field<T>& rhs)
  {
    return field<T>(lhs) -= rhs;
  }

  template<typename T>
  inline field<T> operator * (const field<T>& lhs,
				       const field<T>& rhs)
  {
    return field<T>(lhs) *= rhs;
  }

  template<typename T>
  inline field<T> operator / (const field<T>& lhs,
				       const field<T>& rhs)
  {
    return field<T>(lhs) /= rhs;
  }

  template<typename T>
  inline field<T> operator % (const field<T>& lhs,
				       const field<T>& rhs)
  {
    return field<T>(lhs) %= rhs;
  }

  template<typename T>
  inline bool operator == (const field<T>& lhs,
			   const field<T>& rhs)
  {
    return (lhs.m_value == rhs.m_value);
  }

  template<typename T>
  inline bool operator != (const field<T>& lhs,
			   const field<T>& rhs)
  {
    return !(lhs == rhs);
  }

  template<typename T>
  inline field<T> operator - (const field<T>& rhs)
  {
    field<T> ret(rhs);
    ret.m_value = -ret.m_value;
    return ret;
  }

  template<typename T>
  inline field<T> operator + (const field<T>& rhs)
  {
    field<T> ret(rhs);
    ret.m_value = +ret.m_value;
    return ret;
  }

  template<typename T>
  std::ostream& operator << (std::ostream& os, const field<T>& rhs)
  {
    if (rhs.is_not_null())
      os << *rhs;
    else
      os << "NULL";
    return os;
  }

//   template<typename T>
//   std::istream& operator >> (std::istream& is,
// 			     field<T>& rhs)
//   {
//     T tmp;
//     is >> tmp;
//     if (is)
//       set_value(tmp);
//     return is;
//   }


}; // namespace pqxxobject

#endif // PQXXOBJECT_FIELD_H
