// database field base class                                     -*- C++ -*-
// $Id: field_base.h,v 1.1 2004/05/22 17:58:07 roger Exp $
//
// Copyright (C) 2003  Roger Leigh <rleigh@debian.org>
//
//
//  All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions
//  are met:
//
//  * Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//  * Redistributions in binary form must reproduce the above
//    copyright notice, this list of conditions and the following
//    disclaimer in the documentation and/or other materials provided
//    with the distribution.
//  * Neither the name of the author, nor the names of other
//    contributors may be used to endorse or promote products derived
//    from this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
//  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
//  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
//  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS
//  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
//  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
//  ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
//  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
//  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
//  SUCH DAMAGE.
//
///////////////////////////////////////////////////////////////////////////

#ifndef PQXXOBJECT_FIELD_BASE_H
#define PQXXOBJECT_FIELD_BASE_H

#include <string>

#include <sigc++/object.h>
#include <sigc++/signal.h>

namespace pqxxobject
{
  /**
   * Database field base class.
   * This class is used to represent a single field in a row of a
   * table.  This is a single value belonging to a column in a row,
   * rather than the whole column.  Like a database field, the value
   * may be NULL.  If set to NULL, it is important not to call any of
   * the object member functions, or access the object itself (this
   * would dereference a NULL pointer).
   *
   * The class has some similarity with std::auto_ptr, but the copying
   * semantics are different: on copy, the value is copied, which may
   * involve memory allocation.  There is no transfer of pointer
   * ownership.  Contained objects must posess a copy constructor.
   *
   * As well as storing value, the class has the ability to emit
   * signals when the field value is changed.  Listeners (e.g. user
   * interface widgets) may connect to the signal and will receive
   * notification of changes as they occur.
   */
  class field_base : public SigC::Object
  {
  public:
    /// The constructor.
    field_base();

    /**
     * The constructor.
     * @param column the name of the table column the field belongs
     * to.
     */
    field_base(const std::string& column_name);

    /// The copy constructor.
    field_base(const field_base& rhs);

    /// The destructor.
    virtual ~field_base();

    /**
     * Overloaded assignment operator.
     * The field column name is not copied by the assignment.  This is
     * because differently-named columns could potentially be assigned
     * to one another.
     */
    field_base& operator = (const field_base& rhs);

    /**
     * Get the column name this field belongs to.
     * @returns the column name.
     */
    const std::string& get_column_name() const;

    /**
     * Set the column name this field belongs to.
     * @param column the column name.
     */
    void set_column_name(const std::string& column);

    /**
     * Check if the contained value is NULL.
     * @returns true if NULL, otherwise false.
     */
    virtual bool is_null() const = 0;

    /**
     * Check if the contained value is not NULL.
     * @returns true if not NULL, otherwise false.
     */
    virtual bool is_not_null() const = 0;

    /**
     * Signal emitted on value change.
     * @returns the signal.
     *
     * For example:
     * @code
     * field<int> col;
     * someclass listener;
     * col.signal_changed().connect
     *   ( SigC::slot(listener, &someclass::on_col_changed() );
     * @endcode
     * i.e. a class method (listener.on_col_changed()) will be called
     * when the value is changed.
     */
    SigC::Signal0<void>& signal_changed();

  protected:
    /// The column name.
    std::string m_column_name;

    /// The changed signal.
    SigC::Signal0<void> m_signal_changed;

  }; // class field_base


}; // namespace pqxxobject

#endif // PQXXOBJECT_FIELD_BASE_H
