// database field container proxy                                -*- C++ -*-
// $Id: field_proxy.h,v 1.1 2004/05/22 17:58:08 roger Exp $
//
// Copyright (C) 2004  Roger Leigh <rleigh@debian.org>
//
//
//  All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions
//  are met:
//
//  * Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//  * Redistributions in binary form must reproduce the above
//    copyright notice, this list of conditions and the following
//    disclaimer in the documentation and/or other materials provided
//    with the distribution.
//  * Neither the name of the author, nor the names of other
//    contributors may be used to endorse or promote products derived
//    from this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
//  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
//  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
//  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS
//  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
//  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
//  ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
//  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
//  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
//  SUCH DAMAGE.
//
///////////////////////////////////////////////////////////////////////////

#ifndef PQXXOBJECT_FIELD_PROXY_H
#define PQXXOBJECT_FIELD_PROXY_H

#include <cassert>

#include <sigc++/object.h>
#include <sigc++/signal.h>

#include <pqxxobject/field_base.h>

namespace pqxxobject
{
  /**
   * Database field template proxy class.
   * This class is a proxy for a field class.  That is, it represents
   * a field class, and forwards a subset of the field operations to
   * the referenced field.  It is used to limit access to the field.
   */
  template< typename T >
  class field_proxy
  {
  public:
    typedef typename T::value_type value_type;

    /// The constructor.
    explicit field_proxy(T& value):
      m_value(value)
    {
    }

    /// The copy constructor.
    field_proxy(const field_proxy& rhs):
      m_value(rhs.m_value)
    {
    }

    /// The destructor (non-virtual)
    ~field_proxy()
    {}

    /// Overloaded assignment operator.
    field_proxy<T>& operator = (const value_type& rhs)
    {
      m_value = rhs;
      return *this;
    }

    /// Conversion operator.
    operator value_type () const
    {
      return *m_value;
    }

    /**
     * Check if the contained value is NULL.
     * @returns true if NULL, otherwise false.
     */
    bool is_null() const
    {
      return m_value.is_null();
    }

    /**
     * Check if the contained value is not NULL.
     * @returns true if not NULL, otherwise false.
     */
    bool is_not_null() const
    {
      return m_value.is_not_null();
    }

  private:
    /// The contained value.
    T& m_value;

  }; // class field_proxy

  template< typename T >
  class field_proxy_readonly : public field_proxy<T>
  {
  public:
    typedef typename field_proxy<T>::value_type value_type;

    /// The constructor.
    explicit field_proxy_readonly(T& value):
      field_proxy<T>(value)
    {
    }

    field_proxy_readonly(const field_proxy_readonly& value):
      field_proxy<T>(value)
    {
    }

    /// The destructor (non-virtual)
    ~field_proxy_readonly()
    {}

  private:
    /// Overloaded assignment operator.
    field_proxy_readonly<T>& operator = (const value_type& rhs)
    {
      m_value = rhs;
      return *this;
    }

  }; // class field_proxy_readonly

  template< typename T >
  class field_proxy_writeonly : public field_proxy<T>
  {
  public:
    typedef typename field_proxy<T>::value_type value_type;

    /// The constructor.
    explicit field_proxy_writeonly(T& value):
      field_proxy<T>(value)
    {
    }

    field_proxy_writeonly(const field_proxy_writeonly& value):
      field_proxy<T>(value)
    {
    }

    /// The destructor (non-virtual)
    ~field_proxy_writeonly()
    {}

  private:
    /// Conversion operator.
    operator value_type () const
    {
      return *m_value;
    }

  }; // class field_proxy_writeonly

}; // namespace pqxxobject

#endif // PQXXOBJECT_FIELD_PROXY_H
