// database pointer field container                              -*- C++ -*-
// $Id: ptr_field.h,v 1.1 2004/05/22 17:58:08 roger Exp $
//
// Copyright (C) 2003  Roger Leigh <rleigh@debian.org>
//
//
//  All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions
//  are met:
//
//  * Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//  * Redistributions in binary form must reproduce the above
//    copyright notice, this list of conditions and the following
//    disclaimer in the documentation and/or other materials provided
//    with the distribution.
//  * Neither the name of the author, nor the names of other
//    contributors may be used to endorse or promote products derived
//    from this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
//  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
//  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
//  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS
//  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
//  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
//  ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
//  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
//  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
//  SUCH DAMAGE.
//
///////////////////////////////////////////////////////////////////////////

#ifndef PQXXOBJECT_PTR_FIELD_H
#define PQXXOBJECT_PTR_FIELD_H

#include <cassert>

#include <sigc++/object.h>
#include <sigc++/signal.h>

#include <pqxxobject/field_base.h>

namespace pqxxobject
{
  /**
   * Database pointer field template class.
   * This class is used to represent a single field in a row of a
   * table.  This is a single value belonging to a column in a row,
   * rather than the whole column.  Like a database field, the value
   * may be NULL.  If set to NULL, it is important not to call any of
   * the object member functions, or access the object itself (this
   * would dereference a NULL pointer).
   *
   * The class has some similarity with std::auto_ptr, but the copying
   * semantics are different: on copy, the value is copied, which may
   * involve memory allocation.  There is no transfer of pointer
   * ownership.  Contained objects must posess a copy constructor.
   *
   * As well as storing value, the class has the ability to emit
   * signals when the field value is changed.  Listeners (e.g. user
   * interface widgets) may connect to the signal and will receive
   * notification of changes as they occur.
   */
  template<typename T>
  class ptr_field : public pqxxobject::field_base
  {
  public:
    typedef T value_type;

    /// The constructor.
    ptr_field():
      m_value(0)
    {}

    /// The constructor.
    explicit ptr_field(const value_type& value):
      m_value(0)
    {
      set_value(value);
    }

    /// The copy constructor.
    ptr_field(const ptr_field<value_type>& rhs):
      field_base(rhs),
      m_value(0)
    {
      if (rhs.m_value)
	set_value(*rhs.m_value);
    }

    virtual ~ptr_field()
    {
      if (m_value)
	delete m_value;
    }

    /// Access member functions.
    const value_type *operator -> () const
    {
      assert (m_value != 0);
      return m_value;
    }

    const value_type& operator * () const
    {
      assert (m_value != 0);
      return *m_value;
    }

    /// Overloaded assignment operator.
    ptr_field<value_type>& operator = (const ptr_field<value_type>& rhs)
    {
      if (rhs.m_value)
	set_value(*rhs.m_value);
      else
	set_null();
      return *this;
    }

    /// Overloaded assignment operator.
    ptr_field<value_type>& operator = (const value_type& rhs)
    {
      set_value(rhs);
      return *this;
    }

    ptr_field<value_type>& operator += (const ptr_field<value_type>& rhs)
    {
      assert (m_value != 0);
      assert (rhs.m_value != 0);
      *m_value += *rhs.m_value;
      m_signal_changed.emit();
      return *this;
    }

    ptr_field<value_type>& operator -= (const ptr_field<value_type>& rhs)
    {
      assert (m_value != 0);
      assert (rhs.m_value != 0);
      *m_value -= *rhs.m_value;
      m_signal_changed.emit();
      return *this;
    }

    ptr_field<value_type>& operator *= (const ptr_field<value_type>& rhs)
    {
      assert (m_value != 0);
      assert (rhs.m_value != 0);
      *m_value *= *rhs.m_value;
      m_signal_changed.emit();
      return *this;
    }

    ptr_field<value_type>& operator /= (const ptr_field<value_type>& rhs)
    {
      assert (m_value != 0);
      assert (rhs.m_value != 0);
      *m_value /= *rhs.m_value;
      m_signal_changed.emit();
      return *this;
    }

    ptr_field<value_type>& operator %= (const ptr_field<value_type>& rhs)
    {
      assert (m_value != 0);
      assert (rhs.m_value != 0);
      *m_value %= *rhs.m_value;
      m_signal_changed.emit();
      return *this;
    }

    ptr_field<value_type>& operator ++ ()
    {
      assert (m_value != 0);
      *m_value++;
      m_signal_changed.emit();
      return *this;
    }

    ptr_field<value_type> operator ++ (int)
    {
      assert (m_value != 0);
      ptr_field<value_type> ret(*this);
      *m_value++;
      m_signal_changed.emit();
      return ret;
    }

    ptr_field<value_type>& operator -- ()
    {
      assert (m_value != 0);
      *m_value--;
      m_signal_changed.emit();
      return *this;
    }

    ptr_field<value_type> operator -- (int)
    {
      assert (m_value != 0);
      ptr_field<value_type> ret(*this);
      *m_value--;
      m_signal_changed.emit();
      return ret;
    }

    friend ptr_field<value_type> operator + <> (const ptr_field<value_type>& lhs,
						const ptr_field<value_type>& rhs);

    friend ptr_field<value_type> operator - <> (const ptr_field<value_type>& lhs,
						const ptr_field<value_type>& rhs);

    friend ptr_field<value_type> operator * <> (const ptr_field<value_type>& lhs,
						const ptr_field<value_type>& rhs);

    friend ptr_field<value_type> operator / <> (const ptr_field<value_type>& lhs,
						const ptr_field<value_type>& rhs);

    friend ptr_field<value_type> operator % <> (const ptr_field<value_type>& lhs,
						const ptr_field<value_type>& rhs);

    friend bool operator == <> (const ptr_field<value_type>& lhs,
				const ptr_field<value_type>& rhs);

    friend bool operator != <> (const ptr_field<value_type>& lhs,
				const ptr_field<value_type>& rhs);

    friend ptr_field<value_type> operator - <> (const ptr_field<value_type>& rhs);

    friend ptr_field<value_type> operator + <> (const ptr_field<value_type>& rhs);

    friend std::ostream& operator << <> (std::ostream& output_stream,
					 const ptr_field<value_type>& rhs);

    /// Conversion operator.
    operator const value_type& () const
    {
      assert (m_value != 0);
      return *m_value;
    }

    /**
     * Get the contained value by reference.
     * @returns a reference to the value.
     */
    value_type& get_value()
    {
      assert (m_value != 0);
      return *m_value;
    }

    /**
     * Get the contained value by constant reference.
     * @returns a constant reference to the value.
     */
    const value_type& get_value() const
    {
      assert (m_value != 0);
      return *m_value;
    }

    /**
     * Set the contained value.
     * @param value the value to set.
     */
    void set_value(const value_type& value)
    {
      if (m_value)
	*m_value = value;
      else
	m_value = new value_type(value);
      m_signal_changed.emit();
    }

    /**
     * Set the contained value to NULL.
     */
    void set_null()
    {
      if (m_value)
	{
	  delete m_value;
	  m_value = 0;
	}
    }

    /**
     * Check if the contained value is NULL.
     * @returns true if NULL, otherwise false.
     */
    bool is_null() const
    {
      return m_value == 0;
    }

    /**
     * Check if the contained value is not NULL.
     * @returns true if not NULL, otherwise false.
     */
    bool is_not_null() const
    {
      return m_value != 0;
    }

  private:
    /// The contained value.
    value_type *m_value;

  }; // class ptr_field

  template<typename T>
  inline ptr_field<T> operator + (const ptr_field<T>& lhs,
				  const ptr_field<T>& rhs)
  {
    assert (lhs.m_value != 0);
    assert (rhs.m_value != 0);
    return ptr_field<T>(lhs) += rhs;
  }

  template<typename T>
  inline ptr_field<T> operator - (const ptr_field<T>& lhs,
				  const ptr_field<T>& rhs)
  {
    assert (lhs.m_value != 0);
    assert (rhs.m_value != 0);
    return ptr_field<T>(lhs) -= rhs;
  }

  template<typename T>
  inline ptr_field<T> operator * (const ptr_field<T>& lhs,
				  const ptr_field<T>& rhs)
  {
    assert (lhs.m_value != 0);
    assert (rhs.m_value != 0);
    return ptr_field<T>(lhs) *= rhs;
  }

  template<typename T>
  inline ptr_field<T> operator / (const ptr_field<T>& lhs,
				  const ptr_field<T>& rhs)
  {
    assert (lhs.m_value != 0);
    assert (rhs.m_value != 0);
    return ptr_field<T>(lhs) /= rhs;
  }

  template<typename T>
  inline ptr_field<T> operator % (const ptr_field<T>& lhs,
				  const ptr_field<T>& rhs)
  {
    assert (lhs.m_value != 0);
    assert (rhs.m_value != 0);
    return ptr_field<T>(lhs) %= rhs;
  }

  template<typename T>
  inline bool operator == (const ptr_field<T>& lhs,
			   const ptr_field<T>& rhs)
  {
    assert (lhs.m_value != 0);
    assert (rhs.m_value != 0);
    return (*lhs.m_value == *rhs.m_value);
  }

  template<typename T>
  inline bool operator != (const ptr_field<T>& lhs,
			   const ptr_field<T>& rhs)
  {
    assert (lhs.m_value != 0);
    assert (rhs.m_value != 0);
    return !(lhs == rhs);
  }

  template<typename T>
  inline ptr_field<T> operator - (const ptr_field<T>& rhs)
  {
    assert (rhs.m_value != 0);
    ptr_field<T> ret(rhs);
    *ret.m_value = -(*ret.m_value);
    return ret;
  }

  template<typename T>
  inline ptr_field<T> operator + (const ptr_field<T>& rhs)
  {
    assert (rhs.m_value != 0);
    ptr_field<T> ret(rhs);
    *ret.m_value = +(*ret.m_value);
    return ret;
  }

  template<typename T>
  std::ostream& operator << (std::ostream& os, const ptr_field<T>& rhs)
  {
    if (rhs.is_not_null())
      os << *rhs;
    else
      os << "NULL";
    return os;
  }

}; // namespace pqxxobject

#endif // PQXXOBJECT_PTR_FIELD_H
