// database row base class                                       -*- C++ -*-
// $Id: row_base.cc,v 1.1 2004/05/22 17:58:10 roger Exp $
//
// Copyright (C) 2003  Roger Leigh <rleigh@debian.org>
//
//
//  All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions
//  are met:
//
//  * Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//  * Redistributions in binary form must reproduce the above
//    copyright notice, this list of conditions and the following
//    disclaimer in the documentation and/or other materials provided
//    with the distribution.
//  * Neither the name of the author, nor the names of other
//    contributors may be used to endorse or promote products derived
//    from this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
//  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
//  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
//  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS
//  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
//  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
//  ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
//  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
//  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
//  SUCH DAMAGE.
//
///////////////////////////////////////////////////////////////////////////

#include <pqxxobject/row_base.h>

using namespace pqxxobject;

namespace
{
  std::string describe_status(row_base::row_state current_state)
  {
    std::string status;
    switch (current_state)
      {
      case row_base::STATE_UNINITIALISED:
	status = "the row is not initialised";
	break;
      case row_base::STATE_INITIALISED:
	status = "the row is initialised";
	break;
      case row_base::STATE_INCONSISTENT:
	status = "the row is in an inconsistent state";
	break;
      default:
	status = "the row is in an unknown state";
      }
    return status;
  }
}

row_base::row_base():
  m_state(STATE_UNINITIALISED),
  m_modified(false)
{
}

row_base::row_base(row_state status,
		   bool modified):
  m_state(status),
  m_modified(modified)
{
}

row_base::row_base(const row_base& rhs):
  SigC::Object(),
  m_state(rhs.m_state),
  m_modified(rhs.m_modified)
{
}

row_base::~row_base()
{
}

row_base&
row_base::operator = (const row_base& rhs)
{
  if (&rhs != this)
    {
      m_state = rhs.m_state;
      m_modified = rhs.m_modified;
    }
  return *this;
}

row_base::row_state
row_base::get_state() const
{
  return m_state;
}

bool
row_base::is_modified() const
{
  return m_modified;
}

void
row_base::refresh(pqxxobject::transaction& tran)
{
  try
    {
      if (m_state != STATE_UNINITIALISED)
	{
	  begin(tran);
	  // Register for special handling on transaction commit/abort.
	  connect_safe_signal_handlers(tran);
	  refresh_impl(tran);
	  unset_checkpoint(tran);
	  m_signal_refreshed.emit();
	}
      else
	{
	  abort(tran);
	  throw DatabaseError("The row could not be refreshed: " + describe_status(m_state));
	}
    }
  catch (const DatabaseError& e)
    {
      abort(tran);
      throw e;
    }
}

void
row_base::insert(pqxxobject::transaction& tran)
{
  try
    {
      if (m_state == STATE_UNINITIALISED)
	{
	  begin(tran);
	  // Register for special handling on transaction commit/abort.
	  connect_signal_handlers(tran);
	  insert_impl(tran);
	  unset_checkpoint(tran);
	  m_modified = false;
	  m_signal_inserted.emit();
	}
      else
	{
	  throw DatabaseError("The row could not be inserted: " + describe_status(m_state));
	}
    }
  catch (const DatabaseError& e)
    {
      abort(tran);
      throw e;
    }
}

void
row_base::update(pqxxobject::transaction& tran)
{
  try
    {
      if (m_state == STATE_INITIALISED)
	{
	  begin(tran);
	  // Register for special handling on transaction commit/abort.
	  connect_signal_handlers(tran);
	  update_impl(tran);
	  unset_checkpoint(tran);
	  m_modified = false;
	  m_signal_updated.emit();
	}
      else
	{
	  throw DatabaseError("The row could not be updated: " + describe_status(m_state));
	}
    }
  catch (const DatabaseError& e)
    {
      abort(tran);
      throw e;
    }
}

void
row_base::autoupdate(pqxxobject::transaction& tran)
{
  if (m_state == STATE_UNINITIALISED)
    insert(tran);
  else
    update(tran);
}


void
row_base::erase(pqxxobject::transaction& tran)
{
  try
    {
      if (m_state == STATE_INITIALISED)
	{
	  begin(tran);
	  // Register for special handling on transaction commit/abort.
	  connect_safe_signal_handlers(tran);
	  unset_checkpoint(tran);
	  erase_impl(tran);
	  m_signal_erased.emit();
	}
      else
	{
	  throw DatabaseError("The row could not be deleted: " + describe_status(m_state));
	}
    }
  catch (const DatabaseError& e)
    {
      abort(tran);
      throw e;
    }
}

void
row_base::abort(pqxxobject::transaction& tran)
{
  abort_impl(tran);
}

void
row_base::commit(pqxxobject::transaction& tran)
{
  commit_impl(tran);
}

SigC::Signal0<void>&
row_base::signal_changed()
{
  return m_signal_changed;
}

SigC::Signal0<void>&
row_base::signal_refreshed()
{
  return m_signal_refreshed;
}

SigC::Signal0<void>&
row_base::signal_inserted()
{
  return m_signal_inserted;
}

SigC::Signal0<void>&
row_base::signal_updated()
{
  return m_signal_updated;
}

SigC::Signal0<void>&
row_base::signal_erased()
{
  return m_signal_erased;
}

void
row_base::raise_changed()
{
  m_signal_changed.emit();
}

void
row_base::raise_refreshed()
{
  m_signal_refreshed.emit();
}

void
row_base::raise_inserted()
{
  m_signal_inserted.emit();
}

void
row_base::raise_updated()
{
  m_signal_updated.emit();
}

void
row_base::raise_erased()
{
  m_signal_erased.emit();
}

void
row_base::begin(pqxxobject::transaction& tran)
{
  begin_impl(tran);
  m_state = STATE_INCONSISTENT;
}

void
row_base::refresh_impl(pqxxobject::transaction& tran)
{
}

void
row_base::insert_impl(pqxxobject::transaction& tran)
{
}

void
row_base::update_impl(pqxxobject::transaction& tran)
{
}

void
row_base::erase_impl(pqxxobject::transaction& tran)
{
}

void
row_base::unset_checkpoint(pqxxobject::transaction& tran)
{
  if (tran.get_checkpoint() == true && get_checkpoint() == true)
    tran.set_checkpoint(false);
}

void
row_base::connect_signal_handlers(pqxxobject::transaction& tran)
{
  tran.signal_commit().connect
    ( SigC::slot(*this, &row_base::commit) );
  tran.signal_abort().connect
    ( SigC::slot(*this, &row_base::abort) );
  tran.signal_refresh().connect
    ( SigC::slot(*this, &row_base::refresh) );
}

void
row_base::connect_safe_signal_handlers(pqxxobject::transaction& tran)
{
  tran.signal_commit().connect
    ( SigC::slot(*this, &row_base::commit) );
  tran.signal_abort().connect
    ( SigC::slot(*this, &row_base::abort) );
  // No refresh signal, to prevent recursion.
}
