// database row base class                                       -*- C++ -*-
// $Id: row_base.h,v 1.1 2004/05/22 17:58:10 roger Exp $
//
// Copyright (C) 2003  Roger Leigh <rleigh@debian.org>
//
//
//  All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions
//  are met:
//
//  * Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//  * Redistributions in binary form must reproduce the above
//    copyright notice, this list of conditions and the following
//    disclaimer in the documentation and/or other materials provided
//    with the distribution.
//  * Neither the name of the author, nor the names of other
//    contributors may be used to endorse or promote products derived
//    from this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
//  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
//  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
//  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS
//  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
//  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
//  ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
//  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
//  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
//  SUCH DAMAGE.
//
///////////////////////////////////////////////////////////////////////////

#ifndef PQXXOBJECT_ROW_BASE_H
#define PQXXOBJECT_ROW_BASE_H

#include <map>

#include <sigc++/object.h>
#include <sigc++/signal.h>
#include <sigc++/object_slot.h>

#include <pqxxobject/field_base.h>
#include <pqxxobject/transaction.h>

namespace pqxxobject
{
  /**
   * Database row base class.
   * This class provides the functionality common to all database rows
   * of all types.  Because it derives from SigC::Object, signals may
   * be connected to it, and all classes which derive from it.
   */
  class row_base : public SigC::Object
  {
  public:
    /// The state of the row
    enum row_state
      {
	/**
	 *The row is uninitialised (either new or deleted), not
	 *existing in the database.
	 */
	STATE_UNINITIALISED,
	/// The row exists in the database.
	STATE_INITIALISED,
	/**
	 * The row exists in the database, but is inconsistent due to
	 * an unsuccessful or partially competed transaction.
	 */
	STATE_INCONSISTENT
      };

  protected:
    typedef std::pair<std::string, field_base *> field_type;
    /// The row fields.
    std::map<std::string, field_base *> m_fields;

    /// The row state.
    row_state m_state;
    /// The row modification state.
    bool m_modified;

    /// The changed signal.
    SigC::Signal0<void> m_signal_changed;
    /// The refreshed signal.
    SigC::Signal0<void> m_signal_refreshed;
    /// The inserted signal.
    SigC::Signal0<void> m_signal_inserted;
    /// The updated signal.
    SigC::Signal0<void> m_signal_updated;
    /// The erased signal.
    SigC::Signal0<void> m_signal_erased;

    /// The constructor.
    row_base();

    /**
     * The constructor.
     * @param status the row status.
     * @param modified true if the row is modified, otherwise false.
     */
    explicit row_base(row_state status,
		      bool modified=false);

    /** The copy constructor.
     * @param rhs the object to copy.
     */
    row_base(const row_base& rhs);

  public:
    /// The destructor.
    virtual ~row_base();

    /**
     * Overloaded assignment operator.
     * @param rhs the object to assign.
     */
    row_base& operator = (const row_base& rhs);

    /**
     * Get the row status.
     * @returns the row status.
     */
    row_state get_state() const;

    /**
     * Get the row modification status.
     * @returns true if the row has been modified, false otherwise.
     */
    bool is_modified() const;

    /**
     * Refresh a row from a table.
     * @param tran the transaction which will be used to execute queries.
     */
    void refresh(pqxxobject::transaction& tran);

    /**
     * Insert a row into a table.
     * @param tran the transaction which will be used to execute queries.
     */
    void insert(pqxxobject::transaction& tran);

    /**
     * Update a row from a table, or insert the row if it is not yet
     * present in the table.  This method uses insert() or an update()
     * depending on whether the row has been inserted previously or
     * not.
     * @param tran the transaction which will be used to execute queries.
     */
    void autoupdate(pqxxobject::transaction& tran);

    /**
     * Update a row from a table.
     * @param tran the transaction which will be used to execute queries.
     */
    void update(pqxxobject::transaction& tran);

    /**
     * Remove a row from a table.
     * @param tran the transaction which will be used to execute queries.
     */
    void erase(pqxxobject::transaction& tran);

    /**
     * Recover from transaction abort.
     * @param tran the transaction which will be used to execute queries.
     */
    void abort(pqxxobject::transaction& tran);

    /**
     * Recover from transaction commit.
     * @param tran the transaction which will be used to execute queries.
     */
    void commit(pqxxobject::transaction& tran);

    /**
     * Signal emitted on row change.
     * @returns the signal.
     */
    SigC::Signal0<void>& signal_changed();

    /**
     * Signal emitted on refresh.
     * @returns the signal.
     */
    SigC::Signal0<void>& signal_refreshed();

    /**
     * Signal emitted on insertion.
     * @returns the signal.
     */
    SigC::Signal0<void>& signal_inserted();

    /**
     * Signal emitted on update.
     * @returns the signal.
     */
    SigC::Signal0<void>& signal_updated();

    /**
     * Signal emitted on deletion.
     * @returns the signal.
     */
    SigC::Signal0<void>& signal_erased();

    /// Raise the changed signal.
    void raise_changed();

    /// Raise the refreshed signal.
    void raise_refreshed();

    /// Raise the inserted signal.
    void raise_inserted();

    /// Raise the updated signal.
    void raise_updated();

    /// Raise the erased signal.
    void raise_erased();


  protected:
    /**
     * Begin a database operation.
     * This method is responsible for setting row state and checkpoint
     * objects.
     */
    void begin(pqxxobject::transaction& tran);

    /**
     * Refresh a row from a table.
     * This default method does nothing and should be overridden by
     * derived classes.
     * @param tran the transaction which will be used to execute queries.
     */
    virtual void refresh_impl(pqxxobject::transaction& tran);

    /**
     * Insert a row into a table.
     * This default method does nothing and should be overridden by
     * derived classes.
     * @param tran the transaction which will be used to execute queries.
     */
    virtual void insert_impl(pqxxobject::transaction& tran);

    /**
     * Update a row from a table.
     * This default method does nothing and should be overridden by
     * derived classes.
     * @param tran the transaction which will be used to execute queries.
     */
    virtual void update_impl(pqxxobject::transaction& tran);

    /**
     * Remove a row from a table.
     * This default method does nothing and should be overridden by
     * derived classes.
     * @param tran the transaction which will be used to execute queries.
     */
    virtual void erase_impl(pqxxobject::transaction& tran);

    /**
     * Begin a transaction (checkpoint).
     * This default method does nothing and should be overridden by
     * derived classes.
     * @param tran the transaction which will be used to execute queries.
     */
    virtual void begin_impl(pqxxobject::transaction& tran) = 0;

    /**
     * Recover from transaction abort (roll back to checkpoint).
     * This default method does nothing and should be overridden by
     * derived classes.
     * @param tran the transaction which will be used to execute queries.
     */
    virtual void abort_impl(pqxxobject::transaction& tran) = 0;

    /**
     * Recover from transaction commit (refresh or roll back to checkpoint).
     * This default method does nothing and should be overridden by
     * derived classes.
     * @param tran the transaction which will be used to execute queries.
     */
    virtual void commit_impl(pqxxobject::transaction& tran) = 0;

    /**
     * Find if a checkpoint object exists for this row.
     * @returns true if a checkpoint object exists.
     */
    virtual bool get_checkpoint() const = 0;

    /**
     * Check if the checkpoint status (in the transaction object)
     * needs resetting.
     */
    void unset_checkpoint(pqxxobject::transaction& tran);

  private:
    /**
     * Connect default signal handlers with the transaction.
     * @param tran the transaction which will be used to execute queries.
     */
    void connect_signal_handlers(pqxxobject::transaction& tran);

    /**
     * Connect default signal handlers with the transaction.
     * Only "safe" signal handlers are connected (not signal_refresh)
     * to prevent infinite loops during object refresh.
     * @param tran the transaction which will be used to execute queries.
     */
    void connect_safe_signal_handlers(pqxxobject::transaction& tran);

  }; // class row_base

}; // namespace pqxxobject

#endif // PQXXOBJECT_ROW_BASE_H
