// database transaction convenience wrapper                      -*- C++ -*-
// $Id: transaction.h,v 1.1 2004/05/22 17:58:10 roger Exp $
//
// Copyright (C) 2003  Roger Leigh <rleigh@debian.org>
//
//
//  All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions
//  are met:
//
//  * Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//  * Redistributions in binary form must reproduce the above
//    copyright notice, this list of conditions and the following
//    disclaimer in the documentation and/or other materials provided
//    with the distribution.
//  * Neither the name of the author, nor the names of other
//    contributors may be used to endorse or promote products derived
//    from this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
//  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
//  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
//  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS
//  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
//  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
//  ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
//  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
//  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
//  SUCH DAMAGE.
//
///////////////////////////////////////////////////////////////////////////

#ifndef PQXXOBJECT_TRANSACTION_H
#define PQXXOBJECT_TRANSACTION_H

#include <pqxx/connection>
#include <pqxx/transaction>

#include <sigc++/signal.h>

#include <pqxxobject/exceptions.h>
#include <pqxxobject/query.h>

namespace pqxxobject
{
  /**
   * Database transaction convenience wrapper.
   * This class provides methods to simplify database access for
   * classes which perform any sort of database interaction.  Methods
   * are provided to begin and end transactions, and commit and abort
   * transactions.
   *
   * The transaction class supports recursion.  This means that the
   * transaction object may be passed to functions and class methods
   * other than the owner.  Each user of the transaction object must
   * ensure that they use the object correctly:
   * @code
   * void function(transaction& tran)
   * {
   *   bool error;
   *   tran.begin("function");
   * 
   *   [...]
   * 
   *   if (error)
   *     tran.commit();
   *   else
   *     tran.abort();
   * }
   * @endcode
   * Each function must start by calling the begin() method and end by
   * calling the commit() or abort() method.  The number of begin()
   * calls must match the number of commit() or abort() calls exactly.
   * These are used to keep track of the recursion depth; the commit
   * operation will only take effect at the top level, but an abort
   * will cause the whole transaction to be aborted at all levels.
   */
  class transaction
  {
  public:
    /// The transaction state.
    enum state
      {
	/// No transaction is in progress.
	STATE_NONE,
	/// The current transaction is aborted.
	STATE_ABORTED,
	/// The current transaction is committed.
	STATE_COMMITTED,
	/// The current transaction is executing.
	STATE_EXECUTING,
	/// The transaction is in the process of aborting.
	STATE_ABORTING,
	/// The transaction is in the process of committing.
	STATE_COMMITTING
      };

    /**
     * The constructor.
     * @param conn a reference to a PostgreSQL database
     * connection object.
     * @param tran a pointer to a PostgreSQL transaction
     * object.  It should initially be set to NULL.
     */
    explicit transaction(pqxx::connection& conn,
			 pqxx::transaction<>*& tran);


    /// The destructor.
    virtual ~transaction();

    /**
     * Begin a new transaction.
     * This method should be called at the start of any database
     * operation.  If a transaction does not yet exist it will be
     * created (when called at the top level), otherwise this method
     * has no effect other than counting the recursion depth.
     * @param name the name to give the transaction.  This may be zero
     * length, but will be shown on error messages relating to the
     * transaction.
     */
    void begin(const std::string& name);

  protected:
    /**
     * End a transaction.
     * This method should be called at the end of any database
     * operation.  The transaction will be deleted (when called at the
     * top level), otherwise this method has no effect other than
     * counting the recursion depth.
     * This method has no effect when not running in auto-commit mode.
     */
    void end();

  public:
    /**
     * Execute a query.
     * This method should be called to run an SQL statement on the
     * database server.
     * @param query the SQL statement to run.
     * @returns the result set.
     */
    pqxx::result exec(const std::string& query);

    /**
     * Execute a query.
     * This method should be called to run an SQL statement on the
     * database server.
     * @param query the SQL statement to run.
     * @returns the result set.
     */
    pqxx::result exec(const std::ostringstream& query);

    /**
     * Execute a query.
     * This method should be called to run an SQL statement on the
     * database server.
     * @param query the SQL statement to run.
     * @returns the result set.
     */
    pqxx::result exec(const query& query);

    /**
     * Execute a query without seeing the result set.
     * This method should be called to run an SQL statement on the
     * database server.  Use exec_query if access to the result set is
     * required.
     * @param query the SQL statement to run.
     * @returns the number of rows the query affected.
     */
    pqxx::result::size_type exec_noresult(const std::string& query);

    /**
     * Execute a query without seeing the result set.
     * This method should be called to run an SQL statement on the
     * database server.  Use exec_query if access to the result set is
     * required.
     * @param query the SQL statement to run.
     * @returns the number of rows the query affected.
     */
    pqxx::result::size_type exec_noresult(const std::ostringstream& query);

    /**
     * Execute a query without seeing the result set.
     * This method should be called to run an SQL statement on the
     * database server.  Use exec_query if access to the result set is
     * required.
     * @param query the SQL statement to run.
     * @returns the number of rows the query affected.
     */
    pqxx::result::size_type exec_noresult(const query& query);

    /**
     * Execute a query and automatically commit or abort the changes.
     * This method should be called to run an SQL statement on the
     * database server.  Unlike exec_query(), this method frees the
     * result set automatically.  Use exec_query if access to the
     * result set is required.  The transaction will be committed or
     * aborted depending on if the number of affected rows fell in the
     * specified range.  Note that this is the same as if commit() or
     * abort() were called manually, so the transaction will not
     * actually be committed.  commit() or abort() must still be
     * called to end the transaction block.
     * @param query the SQL statement to run.
     * @param min_rows the minimum number of rows to be altered.  If
     * the number of rows is greater than or equal to this number, the
     * transaction is committed or else it is aborted.  If set to 0,
     * the operation will always be committed.
     * @param max_rows the maximum number of rows to be altered.  If
     * the number of rows is less that or equal to this number, the
     * transaction is committed or else it is aborted.  If set to 0,
     * the operation will always be committed.
     * @returns the number of rows the query affected.
     */
    pqxx::result::size_type perform(const std::string& query,
				    pqxx::result::size_type min_rows,
				    pqxx::result::size_type max_rows);

    /**
     * Execute a query and automatically commit or abort the changes.
     * This method should be called to run an SQL statement on the
     * database server.  Unlike exec_query(), this method frees the
     * result set automatically.  Use exec_query if access to the
     * result set is required.  The transaction will be committed or
     * aborted depending on if the number of affected rows fell in the
     * specified range.  Note that this is the same as if commit() or
     * abort() were called manually, so the transaction will not
     * actually be committed.  commit() or abort() must still be
     * called to end the transaction block.
     * @param query the SQL statement to run.
     * @param min_rows the minimum number of rows to be altered.  If
     * the number of rows is greater than or equal to this number, the
     * transaction is committed or else it is aborted.  If set to 0,
     * the operation will always be committed.
     * @param max_rows the maximum number of rows to be altered.  If
     * the number of rows is less that or equal to this number, the
     * transaction is committed or else it is aborted.  If set to 0,
     * the operation will always be committed.
     * @returns the number of rows the query affected.
     */
    pqxx::result::size_type perform(const std::ostringstream& query,
				    pqxx::result::size_type min_rows,
				    pqxx::result::size_type max_rows);

    /**
     * Execute a query and automatically commit or abort the changes.
     * This method should be called to run an SQL statement on the
     * database server.  Unlike exec_query(), this method frees the
     * result set automatically.  Use exec_query if access to the
     * result set is required.  The transaction will be committed or
     * aborted depending on if the number of affected rows fell in the
     * specified range.  Note that this is the same as if commit() or
     * abort() were called manually, so the transaction will not
     * actually be committed.  commit() or abort() must still be
     * called to end the transaction block.
     * @param query the SQL statement to run.
     * @param min_rows the minimum number of rows to be altered.  If
     * the number of rows is greater than or equal to this number, the
     * transaction is committed or else it is aborted.  If set to 0,
     * the operation will always be committed.
     * @param max_rows the maximum number of rows to be altered.  If
     * the number of rows is less that or equal to this number, the
     * transaction is committed or else it is aborted.  If set to 0,
     * the operation will always be committed.
     * @returns the number of rows the query affected.
     */
    pqxx::result::size_type perform(const query& query,
				    pqxx::result::size_type min_rows,
				    pqxx::result::size_type max_rows);

    /**
     * Commit a transaction to the database.
     * This method should be called after the successful completion of
     * one, or a series of, successful database operations.  The
     * changes will be committed to the database as one single
     * operation.  After calling this method, the transaction may no
     * longer be used.
     *
     * This method only commits the transaction when called at the top
     * level, otherwise it has no effect.
     *
     * On commit, the signal_commit signal is emitted.
     */
    void commit();


    /**
     * Abort a transaction.
     * This method should be called after an error, to revert any
     * changes.  The changes will be rolled back to their state at the
     * start of the transaction as a single operation.  After calling
     * this method, the transaction may no longer be used.
     *
     * On abort, the signal_abort signal is emitted.
     */
    void abort();

    /**
     * Set the transaction checkpoint status.
     * The checkpoint status should be set to true immediately after a
     * checkpoint object has been created, and set back to false
     * immediately after control returns to the function that set it
     * to true in the first place.  This should be taken care of by
     * row_base.
     * @param checkpoint_exists set to true if a checkpoint object
     * exists, otherwise false.
     */
    void set_checkpoint(bool checkpoint_exists = true);

    /**
     * Get the transaction checkpoint status.
     * @returns true if a checkpoint object exists, otherwise false.
     */
    bool get_checkpoint() const;

    /**
     * Signal emitted on transaction commit.
     * All rows that are affected by a transaction are registered with
     * this signal automatically, and their row_base::commit will be
     * called on transaction completion.
     * @returns the signal.
     */
    SigC::Signal1<void, pqxxobject::transaction&>& signal_commit();

    /**
     * Signal emitted on transaction abort.
     * All rows that are affected by a transaction are registered with
     * this signal automatically, and their row_base::abort will be
     * called on transaction completion.
     * @returns the signal.
     */
    SigC::Signal1<void, pqxxobject::transaction&>& signal_abort();

    /**
     * Signal emitted on transaction commit or abort.
     * All rows that are affected by a transaction are registered with
     * this signal automatically, and their row_base::refresh will be
     * called on transaction completion.
     * @returns the signal.
     */
    SigC::Signal1<void, pqxxobject::transaction&>& signal_refresh();

  protected:
    /**
     * Begin a transaction.
     * This method does the actual work.
     * @param name the name to give the transaction.  This may be zero
     * length, but will be shown on error messages relating to the
     * transaction.
     */
    void _begin(const std::string& name);

    /**
     * End a transaction.
     * This method does the actual work.
     */
    void _end();

    /**
     * Commit a transaction.
     * This method is just a stub for _commit().
     * @param refresh true to refresh affected rows on commit, otherwise false.
     */
    void commit(bool refresh);

    /**
     * Commit a transaction.
     * This method does the actual work.
     * @param refresh true to refresh affected rows on commit, otherwise false.
     */
    void _commit(bool refresh);

    /**
     * Abort a transaction.
     * This method is just a stub for _abort().
     * @param rollback true to rollback affected rows on abort, otherwise false.
     */
    void abort(bool rollback);

    /**
     * Abort a transaction.
     * This method does the actual work.
     * @param rollback true to rollback affected rows on abort, otherwise false.
     */
    void _abort(bool rollback);

    // PostgreSQL connection details
    /// A reference to a PostgreSQL database connection object.
    pqxx::connection& m_connection;
    /// A reference to a pointer to a PostgreSQL transaction object.
    pqxx::transaction<>*& m_transaction;

    /// The commit signal.
    SigC::Signal1<void, pqxxobject::transaction&> m_signal_commit;
    /// The abort signal.
    SigC::Signal1<void, pqxxobject::transaction&> m_signal_abort;
    /// The refresh signal.
    SigC::Signal1<void, pqxxobject::transaction&> m_signal_refresh;

    /// The transaction state.
    state m_state;

    /**
     * The recursion level (depth).
     * 0 means no transaction, 1 is the top level and > 1 are
     * recursive transactions.
     */
    unsigned int m_recursion_level;

    /**
     * The checkpoint state.
     * Set to true if a checkpoint object exists for the current
     * object heirarchy, otherwise false.
     */
    bool m_checkpoint;

  }; // class transaction

}; // namespace pqxxobject

#endif // PQXXOBJECT_TRANSACTION_H
