// database update query class                                   -*- C++ -*-
// $Id: update_query.h,v 1.1 2004/05/22 17:58:11 roger Exp $
//
// Copyright (C) 2003  Roger Leigh <rleigh@debian.org>
//
//
//  All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions
//  are met:
//
//  * Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//  * Redistributions in binary form must reproduce the above
//    copyright notice, this list of conditions and the following
//    disclaimer in the documentation and/or other materials provided
//    with the distribution.
//  * Neither the name of the author, nor the names of other
//    contributors may be used to endorse or promote products derived
//    from this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
//  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
//  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
//  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS
//  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
//  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
//  ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
//  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
//  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
//  SUCH DAMAGE.
//
///////////////////////////////////////////////////////////////////////////

#ifndef PQXXOBJECT_UPDATE_QUERY_H
#define PQXXOBJECT_UPDATE_QUERY_H

#include <list>
#include <string>
#include <sstream>
#include <utility>

#include <pqxx/util>

#include <pqxxobject/field_base.h>
#include <pqxxobject/query.h>
#include <pqxxobject/query_helpers.h>

namespace pqxxobject
{
  /**
   * Database UPDATE query class.
   * Convenience class to generate a database UPDATE query string.
   */
  class update_query : public query
  {
  public:
    /// The constructor.
    update_query();

    /**
     * The constructor.
     * @param table the name of the table to update into
     */
    explicit update_query(const std::string& table);

    /// The copy constructor.
    update_query(const update_query& rhs);

    /// The destructor.
    virtual ~update_query();

    typedef std::pair<std::string, std::string> field_value;

    /**
     * Add a field to the query.
     * @param field the field to add.
     */
    template< typename T, template <typename> class F >
    void add(F<T>& field)
    {
      remove(field);
      m_field_list.push_back(query_helper<T, F>::apply(field));
    }

    /**
     * Add a WHERE clause to the query.
     * @param field the field to add.
     */
    template< typename T, template <typename> class F >
    void where(F<T>& field)
    {
      m_where.clear();
      query_helper<T, F> helper;
      std::pair<std::string, std::string> values(helper.apply(field));
      std::ostringstream tmp;
      tmp << values.first << " = " << values.second;
      m_where = tmp.str();
    }


    /**
     * Add a WHERE clause to the query.
     * @param where_string the text of the WHERE clause.
     */
    void where(const std::string& where_string)
    {
      m_where = where_string;
    }

    /**
     * Remove a field from the query.
     * @param field the field to remove.
     */
    void remove(const field_base& field);

    /**
     * Get the query as a string.
     * @returns the query.
     */
    std::string str() const;

  private:
    std::string m_table_name;
    std::list<field_value> m_field_list;
    std::string m_where;

  }; // class update_query

  template<typename T>
  std::ostream& operator << (std::ostream& os, const update_query& rhs)
  {
    os << rhs.str();
  }

}; // namespace pqxxobject

#endif // PQXXOBJECT_UPDATE_QUERY_H
