// tutorial places class                                         -*- C++ -*-
// $Id: places.cc,v 1.12 2004/05/22 17:58:17 roger Exp $
//
// Copyright (C) 2003  Roger Leigh <rleigh@debian.org>
//
//
//  All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions
//  are met:
//
//  * Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//  * Redistributions in binary form must reproduce the above
//    copyright notice, this list of conditions and the following
//    disclaimer in the documentation and/or other materials provided
//    with the distribution.
//  * Neither the name of the author, nor the names of other
//    contributors may be used to endorse or promote products derived
//    from this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
//  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
//  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
//  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS
//  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
//  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
//  ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
//  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
//  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
//  SUCH DAMAGE.
//
///////////////////////////////////////////////////////////////////////////

#include <iostream>
#include <sstream>

#include <pqxxobject/exceptions.h>
#include <pqxxobject/insert_query.h>
#include <pqxxobject/update_query.h>

#include "places.h"

using std::ostringstream;
using std::string;
using std::vector;

using namespace pqxxobject;

Place::Place():
  m_id(0),
  m_name(""),
  m_gridref("")
{
  m_id.set_column_name("id");
  m_name.set_column_name("name");
  m_gridref.set_column_name("gridref");
}

Place::Place(const std::string& name,
	     const std::string& gridref):
  m_id(0),
  m_name(name),
  m_gridref(gridref)
{
  m_id.set_column_name("id");
  m_name.set_column_name("name");
  m_gridref.set_column_name("gridref");
}

Place::~Place()
{
}

int
Place::get_id() const
{
  return m_id;
}

pqxxobject::field_proxy_readonly< pqxxobject::field<int> >
Place::field_id()
{
  return pqxxobject::field_proxy_readonly< pqxxobject::field<int> >(m_id);
}

const std::string&
Place::get_name() const
{
  return m_name;
}

void
Place::set_name(const std::string& name)
{
  m_name = name;
}

pqxxobject::field_proxy< pqxxobject::field<std::string> >
Place::field_name()
{
  return pqxxobject::field_proxy< pqxxobject::field<std::string> >(m_name);
}

const std::string&
Place::get_gridref() const
{
  return m_gridref;
}

void
Place::set_gridref(const std::string& gridref)
{
  m_gridref = gridref;
}

pqxxobject::field_proxy< pqxxobject::field<std::string> >
Place::field_gridref()
{
  return pqxxobject::field_proxy< pqxxobject::field<std::string> >(m_gridref);
}

void
Place::insert_impl(pqxxobject::transaction& tran)
{
//   std::cerr << "Place::insert_impl method called\n";

  tran.begin("Place::insert_impl");

  pqxx::result R(tran.exec("SELECT nextval('places_id_seq') AS sequence"));
  if (R.size() == 1)
    R.begin()->at("sequence").to(m_id.get_value());
  else
    {
      ostringstream reason;
      reason << "Place insertion failed.\n"
             << "Failed to get next place sequence number.\n";
      throw pqxxobject::DatabaseError(reason.str());
    }

  pqxxobject::insert_query query("places");
  query.add(m_id);
  query.add(m_name);
  query.add(m_gridref);

  tran.perform(query, 1, 1);
  // Don't alter m_state--it's set during commit.
  tran.commit();
}

void
Place::update_impl(pqxxobject::transaction& tran)
{
//   std::cerr << "Place::update_impl method called\n";

  tran.begin("Place::update_impl");
//   ostringstream query;
//   query << "UPDATE places "
// 	<< "SET name = '" << get_name()
// 	<< "', gridref = '" << get_gridref() << "' "
// 	<< "WHERE (id = " << get_id() << ")";
  pqxxobject::update_query query("places");
  query.add(m_name);
  query.add(m_gridref);
  query.where(m_id);

  tran.perform(query, 1, 1);
  // Don't alter m_state--it's set during commit.
  tran.commit();
}

void
Place::erase_impl(pqxxobject::transaction& tran)
{
//   std::cerr << "Place::erase_impl method called\n";

  tran.begin("Place::erase_impl");
  ostringstream query;
  query << "DELETE FROM places "
	<< "WHERE (id = " << get_id() << ")";

  tran.perform(query, 1, 1);
  tran.commit();
  // Row is unusable now it's deleted.
  m_id = 0;
  m_state = STATE_UNINITIALISED;
}

void
Place::refresh_impl(pqxxobject::transaction& tran)
{
//   std::cerr << "Place::refresh_impl method called\n";

  tran.begin("Place::refresh_impl");
  ostringstream query;
  query << "SELECT id, name, gridref FROM places "
	<< "WHERE (id = " << get_id() << ")";

  pqxx::result R(tran.exec(query));
  if (R.size() == 1)
    {
      convert_impl(R.begin(), tran);
      // If the refresh worked, we're up-to-date
      m_state = STATE_INITIALISED;
    }
  else // The refresh failed, perhaps the row was deleted?
    {
      m_id = 0;
      m_state = STATE_UNINITIALISED;
    }
  tran.commit();
}

void
Place::convert_impl(pqxx::result::const_iterator row,
		    pqxxobject::transaction& tran)
{
  row->at("id").to(m_id.get_value());
  row->at("name").to(m_name.get_value());
  row->at("gridref").to(m_gridref.get_value());
}


PlaceTable::PlaceTable(pqxxobject::transaction& tran):
  table_base(tran)
{
}

PlaceTable::~PlaceTable()
{
}

PlaceTable::row_list_ptr
PlaceTable::get_list(sort_order order)
{
  ostringstream query;
  query << "SELECT id, name, gridref FROM places "
	<< "ORDER BY ";

  if (order == ORDER_ID)
    query << "id";
  else if (order == ORDER_NAME)
    query << "name";
  else if (order == ORDER_GRIDREF)
    query << "gridref";
  else // fallback
    query << "name";

  return find_many(query.str());
}

PlaceTable::row_ptr
PlaceTable::find(int place_id)
{
  ostringstream query;
  query << "SELECT id, name, gridref FROM places "
	<< "WHERE (id = " << place_id << ")";

  return find_one(query.str());
}


PlaceTable::row_ptr
PlaceTable::find_name(const std::string& name)
{
  ostringstream query;
  query << "SELECT id, name, gridref FROM places "
	<< "WHERE (name = '" << name << "')";

  return find_one(query.str());
}

PlaceTable::row_list_ptr
PlaceTable::find_gridref(const std::string& gridref)
{
  ostringstream query;
  query << "SELECT id, name, gridref FROM places "
	<< "WHERE (gridref = '" << gridref << "')";

  return find_many(query.str());
}
